"""Filtering utils."""

import re
import logging

__all__ = ['left_difference', 'filter_and_sort_annosets']

logger = logging.getLogger(__name__)


def get_text_hash(text):
    """Return hashed text.

    Hash is defined as the text string stripped from all its whitespaces.
    Whitespaces are defined using the regex 's+'.
    """
    return re.sub(r'\s+', '', text.lower().strip())


def _get_sent_hash_set(annosets):
    return {get_text_hash(annoset.sentence.text) for annoset in
            annosets}


def left_difference(source_annosets, target_annosets):
    """Return difference between source and target.

    Remove annosets from source found in target.

    Remove all annosets from source which contain a sentence attribute
    which text hash is in the set of the sentences hash
    of the target annosets
    """
    target_sent_hash_set = _get_sent_hash_set(target_annosets)
    for annoset in source_annosets:
        if get_text_hash(annoset.sentence.text) \
         not in target_sent_hash_set:
            yield annoset
        else:  # TODO: check that we found at least one
            logger.debug('Found hash: {}'.format(get_text_hash(
                annoset.sentence.text)))


def _has_overlapping_fes(annoset):
    if not annoset.labelstore.labels_by_layer_name['FE']:
        return False
    labels_indexes = []
    for label in annoset.labelstore.labels_by_layer_name['FE']:
        if label.start == -1 and label.end == -1:
            continue  # CNI, DNI, INI cases
        for start, end in labels_indexes:
            if label.start >= start and label.start <= end:
                return True
            if label.end >= start and label.end <= end:
                return True
            if label.start <= start and label.end >= end:
                return True
        labels_indexes.append((label.start, label.end))
    return False


def _has_invalid_labels(annoset):
    for label in annoset.labelstore.labels:
        # if the label has an unspecified start/end index
        if label.start == -1 and label.end != -1 \
         or label.start != -1 and label.end == -1:
            return True
        # Deal with some messed-up auto-edited labels in FN 1.7
        if label.start >= len(annoset.sentence.text) \
         or label.end >= len(annoset.sentence.text):
            return True
        # if the label index points at a whitespace
        if (label.start != -1 and annoset.sentence.text[label.start].isspace()) \
         or (label.end != -1 and annoset.sentence.text[label.end].isspace()):
            logger.debug('Found label index pointing at whitespace '
                         'for annoset #{}'.format(annoset._id))
            return True
    return False


def _has_no_fes(annoset):
    if not annoset.labelstore.labels_by_layer_name['FE']:
        return True
    return False


def _has_discontinuous_target(annoset):
    if len(annoset.target.indexes) == 1:
        return False
    prev_end = annoset.target.indexes[0][1]
    for (start, end) in annoset.target.indexes[1:]:
        if start - prev_end > 2:
            return True
        else:
            prev_end = end
    return False


def _has_discontinuous_fes(annoset):
    if not annoset.labelstore.labels_by_layer_name['FE']:
        return False
    # for label in annoset.labelstore.labels_by_layer_name['FE']:
    #     pass  # TODO: implement?
    return False


def _has_non_breaking_spaces(annoset):
    return u'\u00A0' in annoset.sentence.text


# pylint: disable-msg=R0911
def _is_valid_annoset(annoset, filtering_options):
    # No matter what, remove annosets containing invalid labels, i.e. labels
    # with combined specified and unspecified start/end indexes
    if _has_invalid_labels(annoset):
        return False
    # Remove annosets with overlapping frame elements (e.g. for BIOS train)
    if 'overlap_fes' in filtering_options:
        if _has_overlapping_fes(annoset):
            return False
    # Remove annosets with discontinuous frame elements
    if 'disc_fes' in filtering_options:
        if _has_discontinuous_fes(annoset):
            return False
    # Remove annosets with discontinuous targets
    if 'disc_targets' in filtering_options:
        if _has_discontinuous_target(annoset):
            return False
    # Filter annosets with no frame element layers
    if 'no_fes' in filtering_options:
        # the semeval evaluation script will skip sentences with no gold FEs
        if _has_no_fes(annoset):
            return False
    if 'non_breaking_spaces' in filtering_options:
        if _has_non_breaking_spaces(annoset):
            return False
    return True


def _get_target_index_hash(target):
    return '#'.join(['{}#{}'.format(start, end) for (start, end)
                     in target.indexes])


def _get_annoset_target_hash(annoset):
    return '{}#{}#{}'.format(_get_target_index_hash(annoset.target),
                             annoset.target.lexunit.name,
                             annoset.target.lexunit.frame.name)


def _get_annoset_hash(annoset):
    return '{}#{}'.format(get_text_hash(annoset.sentence.text),
                          _get_annoset_target_hash(annoset))


def _filter_annosets(annosets, filtering_options, excluded_frames,
                     excluded_sentences, excluded_annosets):
    annoset_hash_set = set()
    for annoset in annosets:
        if annoset._id in excluded_annosets \
         or annoset.target.lexunit.frame._id in excluded_frames \
         or annoset.sentence._id in excluded_sentences:
            continue
        if not _is_valid_annoset(annoset, filtering_options):
            continue
        annoset_hash = _get_annoset_hash(annoset)
        if annoset_hash in annoset_hash_set:
            continue
        annoset_hash_set.add(annoset_hash)
        yield annoset


def _sort_annosets(annosets):
    """Sort a list of pyfn.AnnotationSet objects.

    Sort by annoset.sentence.text first and then by annoset target hash
    (target.string#target.start#target.end#target.lexunit.name#target.lexunit.frame.name)
    """
    # sort by sentence text as sentences with same text can have different _id
    return sorted(annosets, key=lambda annoset: (annoset.sentence.text,
                                                 _get_annoset_target_hash(
                                                     annoset)))


def filter_and_sort_annosets(annosets, filtering_options, excluded_frames,
                             excluded_sentences, excluded_annosets):
    """Return a generator over a list of sorted and filtered annosets.

    Annosets are filtered according to specified filtering_options.
    Annosets are sorted according to annoset.sentence.text first and
    annoset target hash then:
    (target.string#target.start#target.end#target.lexunit.name#target.lexunit.frame.name)
    Annosets with annoset.target.lexunit.frame.name in excluded_frames are
    excluded.
    Annosets with annoset._id in excluded_annosets are excluded.
    """
    return _sort_annosets(_filter_annosets(annosets, filtering_options,
                                           excluded_frames, excluded_sentences,
                                           excluded_annosets))



"""FrameNet FrameRelation class."""

__all__ = ['FrameRelation']


class FrameRelation():
    """FrameRelation."""

    def __init__(self, _id=None, sub_frame=None, sup_frame=None, frtype=None):
        """Constructor."""
        self.__id = _id
        self._sub_frame = sub_frame
        self._sup_frame = sup_frame
        self._frtype = frtype

    @property
    def _id(self):
        return self.__id

    @property
    def sub_frame(self):
        """Return the subFrame of the FrameRelation."""
        return self._sub_frame

    @property
    def sup_frame(self):
        """Return supFrame of the FrameRelation."""
        return self._sup_frame

    @property
    def frtype(self):
        """Return the FrameRelationType of the FrameRelation."""
        return self._frtype



"""Marshalling to hierarchy files.

Generates ancestors.csv, frame_parent_rolemappings.csv and frame_parents.csv
"""

import itertools
from collections import defaultdict

import logging

import pyfn.utils.files as files_utils
import pyfn.utils.constants as const

__all__ = ['marshall_relations']

logger = logging.getLogger(__name__)


# pylint: disable-msg=R0914
def _get_parents(frame_name, frame_relations_dict):
    parents = []
    if frame_name not in frame_relations_dict:
        return parents
    for parent_name in frame_relations_dict[frame_name]:
        parents.append(parent_name)
    return parents


def _get_frame_parents_dict(frames, frame_relations_dict):
    parents_dict = {}
    for frame_name in sorted(name for name in {frame.name for frame in
                                               frames}):
        parents = _get_parents(frame_name, frame_relations_dict)
        if parents:
            parents_dict[frame_name] = parents
    return parents_dict


def _get_ancestors(frame_name, frame_relations_dict):
    ancestors = []
    if frame_name not in frame_relations_dict:
        return ancestors
    for parent_name in frame_relations_dict[frame_name]:
        ancestors.append(parent_name)
        ancestors.extend(_get_ancestors(parent_name, frame_relations_dict))
    return ancestors


def _get_ancestors_dict(frames, frame_relations_dict):
    ancestors_dict = {}
    for frame_name in sorted(name for name in {frame.name for frame in
                                               frames}):
        ancestors_dict[frame_name] = _get_ancestors(frame_name,
                                                    frame_relations_dict)
    return ancestors_dict


def _get_fe_relations_dict(fe_relations, fe_id_set, relation_type_names):
    fe_relations_dict = defaultdict(list)
    for fe_relation in fe_relations:
        if fe_relation.frame_relation.frtype.name not in relation_type_names:
            continue
        if fe_relation.sub_fe._id not in fe_id_set \
         or fe_relation.sup_fe._id not in fe_id_set:
            continue
        child = '{}.{}'.format(fe_relation.frame_relation.sub_frame.name,
                               fe_relation.sub_fe.name)
        parent = '{}.{}'.format(fe_relation.frame_relation.sup_frame.name,
                                fe_relation.sup_fe.name)
        fe_relations_dict[child].append(parent)
    return fe_relations_dict


def _get_frame_relations_dict(frame_relations, frame_id_set,
                              relation_type_names):
    frame_relations_dict = defaultdict(list)
    for frame_relation in frame_relations:
        if frame_relation.frtype.name not in relation_type_names:
            continue
        if frame_relation.sub_frame._id not in frame_id_set or \
         frame_relation.sup_frame._id not in frame_id_set:
            continue
        frame_relations_dict[frame_relation.sub_frame.name].append(
            frame_relation.sup_frame.name)
    return frame_relations_dict


def _get_fe_id_set(annosets):
    return {label.fe_id for annoset in annosets for label
            in annoset.labelstore.labels_by_layer_name['FE']}


def _get_frame_id_set(annosets):
    return {annoset.target.lexunit.frame._id for annoset in annosets}


def marshall_relations(annosets, frame_relations, fe_relations,
                       target_dirpath):
    """Marshall a relations to csv files."""
    annosets, _annosets, __annosets = itertools.tee(annosets, 3)
    frame_id_set = _get_frame_id_set(annosets)
    frame_relations_dict = _get_frame_relations_dict(
        frame_relations, frame_id_set, const.HIERARCHY_RELATION_TYPES)
    frames = [annoset.target.lexunit.frame for annoset in _annosets]
    ancestors_dict = _get_ancestors_dict(frames, frame_relations_dict)
    ancestors_output_file = files_utils.get_ancestors_filepath(target_dirpath)
    with open(ancestors_output_file, 'w', encoding='utf-8') \
     as ancestors_stream:
        for frame_name, ancestors in ancestors_dict.items():
            if not ancestors:
                print(frame_name, file=ancestors_stream)
            else:
                print('{},{}'.format(frame_name, ','.join(ancestors)),
                      file=ancestors_stream)
    frame_parents_dict = _get_frame_parents_dict(frames, frame_relations_dict)
    frame_parents_output_file = files_utils.get_frame_parents_filepath(
        target_dirpath)
    with open(frame_parents_output_file, 'w', encoding='utf-8') \
     as frame_parents_stream:
        for frame_name, parents in frame_parents_dict.items():
            print('{},{}'.format(frame_name, ','.join(parents)),
                  file=frame_parents_stream)
    fe_id_set = _get_fe_id_set(__annosets)
    fe_relations_dict = _get_fe_relations_dict(
        fe_relations, fe_id_set, const.HIERARCHY_RELATION_TYPES)
    rolemappings_output_file = files_utils.get_rolemappings_filepath(
        target_dirpath)
    with open(rolemappings_output_file, 'w', encoding='utf-8') \
     as rolemappings_stream:
        for role_name, roles in sorted(fe_relations_dict.items()):
            print('{},{}'.format(role_name, ','.join(roles)),
                  file=rolemappings_stream)



"""Storing different formats for labels."""

from collections import defaultdict

__all__ = ['LabelStore']


class LabelStore():
    """Storing lists or dict of labels."""

    def __init__(self, labels):
        """Constructor."""
        self._labels = labels

    @property
    def labels(self):
        """Return a list of Label objects."""
        return self._labels

    @property
    def labels_by_indexes(self):
        """Return a dict of Label objects, excluding PENN, NER and WSL labels.

        Keys are tuples (label.startChar, label.endChar) and values are lists
        of labels
        """
        if not self._labels:
            return {}
        labels_by_indexes = defaultdict(list)
        for label in self._labels:
            labels_by_indexes[(label.start, label.end)].append(label)
        return labels_by_indexes

    @property
    def labels_by_layer_name(self):
        """Return a dict of Label objects, excluding PENN, NER and WSL labels.

        Keys are layer names and values are lists of Label objects
        """
        if not self._labels:
            return {}
        labels_by_layer_name = defaultdict(list)
        for label in self._labels:
            labels_by_layer_name[label.layer.name].append(label)
        return labels_by_layer_name

    @property
    def labels_by_layer(self):
        """Return a dict of Label objects.

        Keys are Layer objects and values are lists of Label objects
        Labelstore in annoset exclude PENN, NER and WSL labels.
        Thoses labels are included in the pnw_labelstore of the Sentence instances
        """
        if not self._labels:
            return {}
        labels_by_layer = defaultdict(list)
        for label in self._labels:
            labels_by_layer[label.layer].append(label)
        return labels_by_layer

    @labels.setter
    def labels(self, labels):
        self._labels = labels



"""FrameNet Layer class."""

__all__ = ['Layer']


class Layer():
    """FrameNet Layer class."""

    def __init__(self, name, rank=None):
        """Constructor."""
        self._name = name
        self._rank = rank

    @property
    def name(self):
        """Return the lexunit name."""
        return self._name

    @property
    def rank(self):
        """Return the layer rank."""
        return self._rank

    @rank.setter
    def rank(self, rank):
        self._rank = rank

    def __hash__(self):
        """Return hash on Layer name and rank."""
        return hash((self._name, self._rank))

    def __eq__(self, other):
        """Compare name and rank fields of two Layer instances."""
        return self._name == other.name and self._rank == other.rank

    def __ne__(self, other):
        """Return standard not equal comparator."""
        return not self == other



"""FrameNet Frame class."""

__all__ = ['Frame']


# pylint: disable=R0903
class Frame():
    """FrameNet Frame class."""

    def __init__(self, name, _id=None):
        """Constructor."""
        self.__id = _id
        self._name = name

    @property
    def _id(self):
        """Return the frame ID."""
        return self.__id

    @property
    def name(self):
        """Return the frame name."""
        return self._name

    @_id.setter
    def _id(self, _id):
        """Set the frame _id."""
        self.__id = _id

    @name.setter
    def name(self, name):
        """Set the frame name."""
        self._name = name



"""(Enriched) FrameNet AnnotationSet object class."""

from pyfn.models.labelstore import LabelStore
from pyfn.models.target import Target
from pyfn.models.valenceunitstore import ValenceUnitStore
from pyfn.models.valencepattern import ValencePattern

__all__ = ['AnnotationSet']


# pylint: disable=R0902
class AnnotationSet():
    """FrameNet AnnotationSet class."""

    # pylint: disable=R0913
    def __init__(self, _id=None, sentence=None, target=None,
                 labelstore=None, vustore=None, valence_pattern=None,
                 c_date=None, xml_schema_type=None):
        """Constructor."""
        self.__id = _id
        self._sentence = sentence
        self._target = target
        self._labelstore = labelstore
        self._vustore = vustore
        self._valence_pattern = valence_pattern
        self._c_date = c_date
        self._xml_schema_type = xml_schema_type

    @classmethod
    def from_fn_data(cls, _id, fn_labels, lexunit, sentence, xml_schema_type,
                     fe_dict, c_date=None):
        """Return an instance of AnnotationSet generated from given FN data."""
        target = Target.from_fn_data(fn_labels, sentence.text,
                                     sentence.pnw_labelstore.labels_by_indexes,
                                     lexunit)
        labelstore = LabelStore(fn_labels)
        vustore = ValenceUnitStore.from_fn_data(
            labelstore.labels_by_indexes, fe_dict)
        valence_pattern = ValencePattern(vustore.valence_units)
        return cls(_id=_id, sentence=sentence, target=target,
                   labelstore=labelstore, vustore=vustore,
                   valence_pattern=valence_pattern, c_date=c_date,
                   xml_schema_type=xml_schema_type)

    @property
    def _id(self):
        """Return the ID of the annotationset."""
        return self.__id

    @property
    def sentence(self):
        """Return the sentence containing the specified annotationset."""
        return self._sentence

    @property
    def target(self):
        """Return the annotated target."""
        return self._target

    @property
    def labelstore(self):
        """Return a LabelStore object."""
        return self._labelstore

    @property
    def vustore(self):
        """Return an ValenceUnitStore object."""
        return self._vustore

    @property
    def valence_pattern(self):
        """Return a ValencePattern object."""
        return self._valence_pattern

    @property
    def c_date(self):
        """Return the annotationset created Date cDate."""
        return self._c_date

    @property
    def xml_schema_type(self):
        """Return the initial XML format.

        The XML format corresponds to the schema from which the AnnotationSet
        was generated: FN 'fulltext', 'exemplar' (lu) or 'semeval'.
        """
        return self._xml_schema_type

    @_id.setter
    def _id(self, _id):
        self.__id = _id

    @target.setter
    def target(self, target):
        self._target = target

    @labelstore.setter
    def labelstore(self, labelstore):
        self._labelstore = labelstore

    @sentence.setter
    def sentence(self, sentence):
        self._sentence = sentence

    @valence_pattern.setter
    def valence_pattern(self, valence_pattern):
        self._valence_pattern = valence_pattern

    def __hash__(self):
        """Return hash on AnnotationSet _id."""
        return hash(self._id)

    def __eq__(self, other):
        """Compare _id fields of two AnnotationSet instances."""
        return (self._id) == (other._id)

    def __ne__(self, other):
        """Return standard not equal comparator."""
        return not self == other



"""Marshaller to BIOS tagging scheme.

BEGINNING = B
INSIDE = I
OUTSIDE = O
SINGULAR = S

BIOS file format:
0   1       2       3               4   5           6       7       8
ID  FORM    LEMMA   PLEMMA(NLTK)    POS PPOS(NLTK)  FEAT    PFEAT   HEAD

9       10      11      12          13      14      15
PHEAD   DEPREL  PDEPREL FILLPRED    PRED    APREDS  FE_CORETYPE
"""

import itertools
import logging

import pyfn.utils.files as files_utils
import pyfn.utils.filter as filt_utils
import pyfn.utils.marshalling as marsh_utils

from pyfn.exceptions.parameter import InvalidParameterError

__all__ = ['marshall_annosets_dict']

logger = logging.getLogger(__name__)


def _get_fe_label(start, end, fe_labels):
    for label in fe_labels:
        # FIXME: careful, this is not compatible with discontinuous labels
        if label.start >= start and label.end <= end:
            return 'S-{}'.format(label.name)
        if label.start == start:
            return 'B-{}'.format(label.name)
        if label.start < start and label.end >= start:
            return 'I-{}'.format(label.name)
    return 'O'


def _get_lexunit(start, end, target):
    if (start, end) in target.indexes:
        return target.lexunit
    for istart, iend in target.indexes:
        if not (iend < start or istart > end):
            return target.lexunit
    return '_'


def _get_frame_name(lexunit):
    return '_' if lexunit == '_' else lexunit.frame.name


def _get_lexunit_name(lexunit):
    return '_' if lexunit == '_' else lexunit.name


def _get_token_lemma(token, pos, lemmatizer):
    if pos.startswith('V'):
        return lemmatizer.lemmatize(token, pos='v')
    return lemmatizer.lemmatize(token)


def _get_token_index_3uples(text):
    token_index_3uples = []
    tokens = text.split()
    start_index = 0
    for token in tokens:
        while token[0] != text[start_index]:
            start_index += 1
        token_index_3uples.append((token, start_index,
                                   start_index+len(token)-1))
        start_index += len(token)
    return token_index_3uples


def _get_valence_units_by_indexes(vus_by_indexes_dict, start, end):
    valence_units = []
    for (vu_start, vu_end), vus in vus_by_indexes_dict.items():
        if not (vu_end < start or vu_start > end):
            valence_units.extend(vus)
    return valence_units


def _get_fe_coretype(fe_label, vus):
    if fe_label == 'O':
        raise InvalidParameterError('Unspecified FE label: {}'.format(fe_label))
    if not vus:
        raise InvalidParameterError('Input ValenceUnit list is empty')
    fe_name = fe_label[2:]
    for vu in vus:
        if vu.fe.name == fe_name:
            return vu.fe.coretype
    raise Exception('Could not output FE coreType: no matching FE name \'{}\' '
                    'found in ValenceUnit list {}'.format(fe_name, vus))


def _get_bios_lines(annoset, sent_dict, with_fe_anno=False):
    bios_lines = []
    token_index_3uples = _get_token_index_3uples(annoset.sentence.text)
    sent_num = marsh_utils.get_sent_num(annoset.sentence.text, sent_dict)
    token_num = 1
    for token_3uple in token_index_3uples:
        if not with_fe_anno or 'FE' not in annoset.labelstore.labels_by_layer_name:
            fe_label = 'O'
            fe_coretype = '_'
        else:
            fe_label = _get_fe_label(
                token_3uple[1], token_3uple[2],
                annoset.labelstore.labels_by_layer_name['FE'])
            if fe_label == 'O':
                fe_coretype = '_'
            else:
                vus = _get_valence_units_by_indexes(annoset.vustore.valence_units_by_indexes,
                                                    token_3uple[1],
                                                    token_3uple[2])
                fe_coretype = _get_fe_coretype(fe_label, vus)
        bios_line = '{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}'.format(
            token_num,
            token_3uple[0],
            '_',
            '_',
            '_',
            '_',
            sent_num,
            '_', '_', '_', '_', '_',
            _get_lexunit_name(_get_lexunit(token_3uple[1], token_3uple[2],
                                           annoset.target)),
            _get_frame_name(_get_lexunit(token_3uple[1], token_3uple[2],
                                         annoset.target)),
            fe_label,
            fe_coretype)
        bios_lines.append(bios_line)
        token_num += 1
    return bios_lines


def _marshall_bios(annosets, filtering_options, sent_dict, bios_filepath,
                   excluded_frames, excluded_sentences, excluded_annosets,
                   with_fe_anno):
    files_utils.create_parent_dir_if_not_exists(bios_filepath)
    with open(bios_filepath, 'w', encoding='utf-8') as bios_stream:
        for annoset in filt_utils.filter_and_sort_annosets(annosets,
                                                           filtering_options,
                                                           excluded_frames,
                                                           excluded_sentences,
                                                           excluded_annosets):
            bios_lines = _get_bios_lines(annoset, sent_dict, with_fe_anno)
            print('\n'.join(bios_lines), file=bios_stream)
            print('', file=bios_stream)  # at the end of a sentence


def marshall_annosets_dict(annosets_dict, target_dirpath, filtering_options,
                           output_sentences, excluded_frames,
                           excluded_sentences, excluded_annosets):
    """Convert a dict of {splits:pyfn.AnnotationSet} to BIOS splits files."""
    for splits_name, annosets in annosets_dict.items():
        bios_filepath = files_utils.get_bios_filepath(target_dirpath,
                                                      splits_name)
        bios_semeval_filepath = files_utils.get_bios_semeval_filepath(
            target_dirpath, splits_name)
        sent_filepath = files_utils.get_sent_filepath(target_dirpath,
                                                      splits_name)
        sent_dict = {}
        if splits_name not in ['train', 'dev', 'test']:
            raise InvalidParameterError('Invalid splits_name: {}'.format(
                splits_name))
        if splits_name == 'dev' or splits_name == 'test':
            annosets, _annosets = itertools.tee(annosets, 2)
            logger.info('Marshalling splits:pyfn.AnnotationSet dict to '
                        '.bios.semeval for {} splits with [] filtering '
                        'options...'.format(splits_name))
            _marshall_bios(annosets, [], sent_dict,  # No special filtering on dev/test
                           bios_semeval_filepath, excluded_frames,
                           excluded_sentences,
                           excluded_annosets, with_fe_anno=False)
            logger.info('Marshalling splits:pyfn.AnnotationSet dict to '
                        '.bios for {} splits with {} filtering '
                        'options...'.format(splits_name, filtering_options))
            _marshall_bios(_annosets, filtering_options, sent_dict,
                           bios_filepath, excluded_frames, excluded_annosets,
                           excluded_sentences, with_fe_anno=True)
        elif splits_name == 'train':
            logger.info('Marshalling splits:pyfn.AnnotationSet dict to '
                        '.bios for {} splits with {} filtering '
                        'options...'.format(splits_name, filtering_options))
            _marshall_bios(annosets, filtering_options, sent_dict,
                           bios_filepath, excluded_frames,
                           excluded_sentences, excluded_annosets,
                           with_fe_anno=True)
        # print out sentences file
        if output_sentences:
            marsh_utils.marshall_sent_dict(sent_dict, sent_filepath)



"""Welcome to pyfn.

This is the entry point of the application.
"""

import os
import argparse

import logging
import logging.config

import pyfn.marshalling.marshallers.bios as biosm
import pyfn.marshalling.marshallers.hierarchy as hierm
import pyfn.marshalling.marshallers.semafor as semaform
import pyfn.marshalling.marshallers.semeval as semeval
import pyfn.marshalling.unmarshallers.bios as biosu
import pyfn.marshalling.unmarshallers.framenet as fnxml
import pyfn.marshalling.unmarshallers.semafor as semaforu

import pyfn.utils.config as config_utils
import pyfn.utils.files as futils

from pyfn.exceptions.parameter import InvalidParameterError

# logging.config.dictConfig(
#     config_utils.load(
#         os.path.join(os.path.dirname(__file__), 'logging', 'logging.yml')))

logger = logging.getLogger(__name__)


def _extract_splits_name(source_path):
    return os.path.splitext(os.path.basename(source_path))[0]


def _generate(args):
    train_annosets = fnxml.get_annosets_dict(args.splits_path, 'train',
                                             args.with_exemplars)['train']
    fr_relation_xml = futils.get_fr_relation_xml_filepath(args.splits_path)
    frame_relations, fe_relations = fnxml.extract_relations(fr_relation_xml)
    hierm.marshall_relations(train_annosets, frame_relations, fe_relations,
                             args.target_path)


def _convert(args):
    if args.source_format == args.target_format:
        raise InvalidParameterError(
            'Source and Target formats are the same! Please specify different '
            'source/target formats')
    if args.source_path == args.target_path:
        raise InvalidParameterError(
            'Source and Target paths are the same! Please specify different '
            'source/target paths')
    # TODO: add validation for input directory structure
    if args.source_format == 'fnxml':
        annosets_dict = fnxml.get_annosets_dict(args.source_path,
                                                args.splits,
                                                args.with_exemplars)
    if args.source_format == 'bios':
        if args.sent == '__undefined__':
            raise InvalidParameterError(
                'Unspecified sentence file. For bios unmarshalling you need '
                'to specify the --sent parameter pointing at the '
                '.sentences file absolute filepath')
        annosets = biosu.unmarshall_annosets(args.source_path, args.sent)
    if args.source_format == 'semafor':
        if args.sent == '__undefined__':
            raise InvalidParameterError(
                'Unspecified sentence file. For semafor unmarshalling you '
                'need to specify the --sent parameter pointing at the '
                '.sentences file absolute filepath')
        annosets = semaforu.unmarshall_annosets(args.source_path, args.sent)
    if args.target_format == 'bios':
        biosm.marshall_annosets_dict(annosets_dict, args.target_path,
                                     args.filter, args.output_sentences,
                                     args.excluded_frames,
                                     args.excluded_sentences,
                                     args.excluded_annosets)
    if args.target_format == 'semeval':
        if args.source_format == 'fnxml':
            splits_name = args.splits
            annosets = annosets_dict[splits_name]
            output_filepath = os.path.join(args.target_path,
                                           '{}.gold.xml'.format(splits_name))
        if args.source_format == 'bios' or args.source_format == 'semafor':
            output_filepath = args.target_path
        semeval.marshall_annosets(annosets, output_filepath,
                                  args.excluded_frames,
                                  args.excluded_sentences,
                                  args.excluded_annosets)
    if args.target_format == 'semafor':
        semaform.marshall_annosets_dict(annosets_dict, args.target_path,
                                        args.filter, args.output_sentences,
                                        args.excluded_frames,
                                        args.excluded_sentences,
                                        args.excluded_annosets)


def main():
    """Launch the pyfn application."""
    parser = argparse.ArgumentParser(prog='pyfn')
    subparsers = parser.add_subparsers()
    parser_generate = subparsers.add_parser(
        'generate', formatter_class=argparse.RawTextHelpFormatter,
        help='generate hierarchy files')
    parser_generate.set_defaults(func=_generate)
    parser_generate.add_argument('--source', required=True,
                                 dest='splits_path',
                                 help='absolute filepath to source dir')
    parser_generate.add_argument('--target', required=True,
                                 dest='target_path',
                                 help='absolute filepath to target dir')
    parser_generate.add_argument('--with_exemplars',
                                 action='store_true', default=False,
                                 help='whether or not to use exemplars in '
                                      'splits. Default to false')

    parser_convert = subparsers.add_parser(
        'convert', formatter_class=argparse.RawTextHelpFormatter,
        help='convert source file from given format to target file in given '
             'format')
    parser_convert.set_defaults(func=_convert)
    parser_convert.add_argument('--source', required=True,
                                dest='source_path',
                                help='absolute filepath to source dir')
    parser_convert.add_argument('--target', required=True,
                                dest='target_path',
                                help='absolute filepath to target dir')
    parser_convert.add_argument('--from', required=True,
                                dest='source_format',
                                choices=['semafor', 'bios', 'semeval',
                                         'fnxml'],
                                help='''source format. Choose between:
    - semafor: the format used by the semafor parser
    - bios: the BIOS format used by the open-sesame parser
    - semeval: the SEMEVAL 2008 XML format
    - fnxml: the standard FrameNet XML format
    ''')
    parser_convert.add_argument('--to', required=True,
                                dest='target_format',
                                choices=['semafor', 'bios', 'semeval',
                                         'fnxml'],
                                help='''target format. Choose between:
    - conll: the CoNLL format used by the semafor parser
    - bios: the BIOS format used by the open-sesame parser
    - semeval: the SEMEVAL 2008 XML format
    - fnxml: the standard FrameNet XML format
    ''')
    parser_convert.add_argument('--with_exemplars',
                                action='store_true', default=False,
                                help='whether or not to use exemplars in '
                                     'splits. Default to false')
    parser_convert.add_argument('--output_sentences',
                                action='store_true', default=False,
                                help='whether or not to output the .sentences '
                                     'files in bios or semafor marshalling')
    parser_convert.add_argument('--splits',
                                choices=['train', 'dev', 'test'],
                                default='test',
                                help='names of FrameNet splits to be '
                                     'unmarshalled')
    parser_convert.add_argument('--sent',
                                default='__undefined__',
                                help='absolute path to the '
                                     '{train,dev,test}.sentences file for '
                                     'BIOS unmarshalling')
    parser_convert.add_argument('--filter',
                                nargs='+',
                                default=[],
                                help='''filtering options for the training set:
    - overlap_fes: filter out annosets with overlapping frame elements
    - disc_fes: filter out annosets with discontinuous frame elements
    - disc_targets: filter out annosets with discontinuous targets
    - no_fes: filter out annotationsets with no frame element labels
    - non_breaking_spaces: filter out annotationsets with sentence.text
      containing non-breaking spaces
    ''')
    parser_convert.add_argument('--excluded_frames',
                                nargs='+',
                                type=int,
                                default=[],
                                help='list of frame ids to be excluded from '
                                     'unmarshalled FrameNet XML')
    parser_convert.add_argument('--excluded_sentences',
                                nargs='+',
                                type=int,
                                default=[],
                                help='list of sentence ids to be excluded '
                                     'from unmarshalled FrameNet XML')
    parser_convert.add_argument('--excluded_annosets',
                                nargs='+',
                                type=int,
                                default=[],
                                help='list of annoset ids to be excluded from '
                                     'unmarshalled FrameNet XML')
    args = parser.parse_args()
    args.func(args)



"""FrameNet Label class."""

__all__ = ['Label']


# pylint: disable=R0902
class Label():
    """FrameNet Label class."""

    # pylint: disable=R0913
    def __init__(self, name=None, layer=None, start=None, end=None, fe_id=None,
                 itype=None, fg_color=None, bg_color=None):
        """Constructor."""
        self._name = name
        self._start = start
        self._end = end
        self._layer = layer
        self._fe_id = fe_id
        self._itype = itype
        self._fg_color = fg_color
        self._bg_color = bg_color

    @property
    def name(self):
        """Return the label name."""
        return self._name

    @property
    def start(self):
        """Return the label start char index."""
        return self._start

    @property
    def end(self):
        """Return the label end char index."""
        return self._end

    @property
    def fe_id(self):
        """Return the label feID (if it is a FrameElement)."""
        return self._fe_id

    @property
    def itype(self):
        """Return the label instantiation type (if set)."""
        return self._itype

    @property
    def fg_color(self):
        """Return the label fgColor (if it is a FrameElement)."""
        return self._fg_color

    @property
    def bg_color(self):
        """Return the label bgColor (if it is a FrameElement)."""
        return self._bg_color

    @property
    def layer(self):
        """Return the label layer."""
        return self._layer

    @start.setter
    def start(self, start):
        self._start = start

    @end.setter
    def end(self, end):
        self._end = end

    @fe_id.setter
    def fe_id(self, fe_id):
        self._fe_id = fe_id

    @itype.setter
    def itype(self, itype):
        self._itype = itype

    @fg_color.setter
    def fg_color(self, fg_color):
        self._fg_color = fg_color

    @bg_color.setter
    def bg_color(self, bg_color):
        self._bg_color = bg_color

    def __str__(self):
        """Return stringified label.

        Return {name}#{layer.name}#{itype} if label is a null
        instantiation (itype).
        Return {name}#{layer.name}#{start}#{end} otherwise.
        """
        if self.itype:
            return '{}#{}#{}'.format(self.name, self.layer.name, self.itype)
        return '{}#{}#{}#{}'.format(self.name, self.layer.name, self.start,
                                    self.end)

    def __repr__(self):
        """Return stringified object."""
        return str(self)



"""Storing valence units in multiple format."""

from collections import defaultdict

import logging

from pyfn.models.frameelement import FrameElement
from pyfn.models.valenceunit import ValenceUnit

__all__ = ['ValenceUnitStore']

logger = logging.getLogger(__name__)


# pylint: disable=R0916
def _contains_unspecified_fe_pt_gf(index, labels):
    """Return true iff.

    at least one label is a FE/PT/GF and not all of them are specified.
    """
    if index[0] != -1 and index[1] != -1:
        contains_fe = False
        contains_pt = False
        contains_gf = False
        for label in labels:
            if label.layer.name == 'FE':
                contains_fe = True
            if label.layer.name == 'PT':
                contains_pt = True
            if label.layer.name == 'GF':
                contains_gf = True
        if (contains_fe or contains_pt or contains_gf) and \
         (not contains_fe or not contains_pt or not contains_gf):
            return True
    return False


def _contains_unspecified_indexes(index, labels):
    if index[0] == -1 or index[1] == -1:  # if start or end is not specified
        for label in labels:
            if label.name == 'PT' or label.name == 'GF':
                logger.warning(
                    'start/end indexes are not specified for PT/GF: {}'
                    .format(label.name))
                return True
            if label.name == 'FE' and label.itype is None:
                logger.warning(
                    'start/end indexes are not specified for FE {}'
                    .format(label.name))
                return True
    return False


# pylint: disable=R0912
def _to_valence_units_by_indexes(labels_by_indexes, fe_dict):
    valence_units_by_indexes = defaultdict(list)
    for index, labels in labels_by_indexes.items():
        if _contains_unspecified_indexes(index, labels):
            valence_units_by_indexes[index] = []
        else:
            if index[0] == -1 or index[1] == -1:  # if start or end is not
                # specified
                for label in labels:
                    if fe_dict:
                        fe = fe_dict[label.fe_id]
                    else:
                        fe = FrameElement(_id=label.fe_id,
                                          name=label.name)
                    fe.itype = label.itype
                    valence_unit = ValenceUnit(fe)
                    valence_units_by_indexes[index].append(valence_unit)
            else:
                fe_labels = [label for label in labels
                             if label.layer.name == 'FE']
                pt_labels = [label for label in labels
                             if label.layer.name == 'PT']
                gf_labels = [label for label in labels
                             if label.layer.name == 'GF']
                if not fe_labels and not pt_labels and not gf_labels:
                    # We are not dealing with FN FE.PT.GF layers
                    continue
                for fe_label in fe_labels:
                    if fe_dict:
                        fe = fe_dict[fe_label.fe_id]
                    else:
                        fe = FrameElement(_id=fe_label.fe_id,
                                          name=fe_label.name)
                    if pt_labels and gf_labels:  # TODO check also this
                        valence_unit = ValenceUnit(fe, pt_labels[0].name,
                                                   gf_labels[0].name)
                        valence_units_by_indexes[index].append(valence_unit)
                    else:
                        valence_unit = ValenceUnit(fe, 'undefined',
                                                   'undefined')
                        valence_units_by_indexes[index].append(valence_unit)
    return valence_units_by_indexes


def _to_valence_units(valence_units_by_indexes):
    all_valence_units = []
    for valence_units in valence_units_by_indexes.values():
        if not valence_units:
            return []  # Strict rule: if there is one problem, return []
        all_valence_units.extend(valence_units)
    return all_valence_units


class ValenceUnitStore():
    """ValenceUnitStore."""

    def __init__(self, valence_units, valence_units_by_indexes):
        """Constructor."""
        self._valence_units = valence_units
        self._valence_units_by_indexes = valence_units_by_indexes

    @classmethod
    def from_fn_data(cls, fn_labels_by_indexes, fe_dict):
        """Return ValenceUnitStore instance generated from FrameNet data."""
        valence_units_by_indexes = _to_valence_units_by_indexes(
            fn_labels_by_indexes, fe_dict)
        valence_units = _to_valence_units(valence_units_by_indexes)
        return cls(valence_units, valence_units_by_indexes)

    @property
    def valence_units_by_indexes(self):
        """Return an index to ValenceUnit objects dict."""
        return self._valence_units_by_indexes

    @property
    def valence_units(self):
        """Return a list of ValenceUnit objects."""
        return self._valence_units



"""FrameNet Valence Unit."""

__all__ = ['ValenceUnit']


# pylint: disable=C0103
class ValenceUnit():
    """ValenceUnit: Frame Element + Phrase Type + Grammatical Function."""

    def __init__(self, fe, pt=None, gf=None):
        """Constructor."""
        self._fe = fe
        self._pt = pt
        self._gf = gf

    @property
    def fe(self):
        """Return a FrameElement object."""
        return self._fe

    @property
    def pt(self):
        """Return a string for the phrase type."""
        return self._pt

    @property
    def gf(self):
        """Return a string for the grammatical function."""
        return self._gf

    @property
    def with_fe_name(self):
        """Return FE.PT.GF with FE as FE name."""
        if self._fe.itype:
            return '{}.{}'.format(self._fe.name, self._fe.itype)
        return '{}.{}.{}'.format(self._fe.name, self._pt, self._gf)

    @property
    def with_fe_cnc(self):
        """Return FE.PT.GF with FE as FE.name#C or FE.name#NC.

        C and NC stand for Core and Non-Core. No additionnal specification
        is given as to the type of Non-Core FE.
        """
        if not self._fe.coretype:
            return ''
        if self._fe.itype:
            if self._fe.coretype == 'Core' \
             or self._fe.coretype == 'Core-Unexpressed':
                return '{}#C.{}'.format(self._fe.name, self._fe.itype)
            return '{}#NC.{}'.format(self._fe.name, self._fe.itype)
        if self._fe.coretype == 'Core' \
         or self._fe.coretype == 'Core-Unexpressed':
            return '{}#C.{}.{}'.format(self._fe.name, self._pt, self._gf)
        return '{}#NC.{}.{}'.format(self._fe.name, self._pt, self._gf)

    @property
    def with_fe_coretype(self):
        """Return FE.PT.GF with FE as FE.name#FE.coreType."""
        if not self._fe.coretype:
            return ''
        if self._fe.itype:
            return '{}#{}.{}'.format(self._fe.name, self._fe.coretype,
                                     self._fe.itype)
        return '{}#{}.{}.{}'.format(self._fe.name, self._fe.coretype,
                                    self._pt, self._gf)



"""Load FN data."""

import os
import pickle
import logging
import mmh3
import pyfn.utils.constants as const
import pyfn.utils.xml as xml_utils

__all__ = ['load_fe_dict']

logger = logging.getLogger(__name__)


def _get_fe_dict_identifier(frame_xml_dirpath):
    identifier = ''
    xml_filepaths = xml_utils.extract(frame_xml_dirpath)
    for xml_filepath in xml_filepaths:
        identifier = '{}{}{}'.format(identifier,
                                     os.path.basename(xml_filepath),
                                     os.path.getmtime(xml_filepath))
    return str(mmh3.hash(identifier))


def load_fe_dict(frame_xml_dirpath):
    """Load a {fe_id: FrameElement} dictionary."""
    identifier = _get_fe_dict_identifier(frame_xml_dirpath)
    fe_dict_filepath = os.path.join(const.DATA_DIR,
                                    '{}.pkl'.format(identifier))
    if os.path.exists(fe_dict_filepath):
        logger.debug('Found fe dict matching config identifier. Loading...')
        with open(fe_dict_filepath, 'rb') as identifier_stream:
            return pickle.load(identifier_stream)
    logger.debug('Could not find fe dict matching identifier:')
    # TODO: implement frame unmarshaller
    # fe_dict = unmarshaller.unmarshall_frame_xml(frame_xml_dirpath)
    fe_dict = {}
    with open(fe_dict, 'wb') as identifier_stream:
        pickle.dump(fe_dict, identifier_stream,
                    protocol=pickle.HIGHEST_PROTOCOL)
    return fe_dict



"""FrameNet Sentence class."""

from pyfn.models.labelstore import LabelStore

__all__ = ['Sentence']


class Sentence():
    """FrameNet sentence class."""

    def __init__(self, _id=None, text=None, pnwb_labels=None, document=None):
        """Constructor."""
        self.__id = _id
        self._document = document
        self._text = text
        self._sentence_number = None
        self._paraphraph_number = None  # TODO : add this and doc/corpID for lexunits
        self._pnw_labelstore = LabelStore(pnwb_labels)

    @property
    def _id(self):
        """Return the ID of the FrameNet sentence."""
        return self.__id

    @property
    def text(self):
        """Return the text of the FramNet sentence."""
        return self._text

    @property
    def document(self):
        """Return the document object containing the FrameNet sentence."""
        return self._document

    @property
    def pnw_labelstore(self):
        """Return a LabelStore for PENN, NER, WSL and BNC labels."""
        return self._pnw_labelstore

    @document.setter
    def document(self, document):
        self._document = document

    @text.setter
    def text(self, text):
        self._text = text



#!/usr/bin/env python3
"""pyfn setup.py.

This file details modalities for packaging the pyfn application.
"""

from setuptools import setup

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='pyfn',
    description='A python module to process FrameNet XML data',
    author='Alexandre Kabbach',
    author_email='akb@3azouz.net',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version='1.0.0rc4',
    url='https://gitlab.unige.ch/akb/pyfn',
    download_url='https://pypi.org/project/pyfn/#files',
    license='MIT',
    keywords=['framenet', 'xml', 'marshalling', 'unmarshalling'],
    platforms=['any'],
    packages=['pyfn',
              'pyfn.exceptions',
              'pyfn.loading',
              'pyfn.marshalling',
              'pyfn.marshalling.marshallers',
              'pyfn.marshalling.unmarshallers',
              'pyfn.models',
              'pyfn.utils'],
    package_data={'pyfn': ['logging/*.yml']},
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'pyfn = pyfn.main:main'
        ],
    },
    tests_require=['pytest==3.4.1', 'pylint==1.8.2', 'pytest-cov==2.5.1',
                   'pydocstyle==2.1.1'],
    install_requires=['PyYAML==3.12', 'mmh3==2.5.1', 'lxml==4.2.1',
                      'pytz==2018.4'],
    classifiers=['Development Status :: 4 - Beta',
                 'Environment :: Console',
                 'Intended Audience :: Developers',
                 'Intended Audience :: Education',
                 'Intended Audience :: Science/Research',
                 'License :: OSI Approved :: MIT License',
                 'Natural Language :: English',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Topic :: Scientific/Engineering :: Artificial Intelligence',
                 'Topic :: Software Development :: Libraries :: Python Modules',
                 'Topic :: Text Processing :: Linguistic'],
    zip_safe=True,
)



"""Config utilities.

Methods used to manipulate YAML-based configuration files.
"""

import logging
import yaml

from pyfn.exceptions.method import InvalidMethodError
from pyfn.exceptions.parameter import InvalidParameterError
from pyfn.utils.immutables import ImmutableConfig

__all__ = ['load']

logger = logging.getLogger(__name__)


def load(config_file):
    """Load an ImmutableConfig from a YAML configuration file."""
    logger.info('Loading config from file {}'.format(config_file))
    try:
        with open(config_file, 'r') as config_stream:
            config = yaml.safe_load(config_stream)
            return ImmutableConfig(config)
    except (InvalidMethodError, InvalidParameterError):
        raise



"""Marshaller to SEMAFOR/SEMAFOR format.

SEMAFOR file format:
0   1       2               3       4       5                   6
1   0.0     #(Frame + FEs)  Frame   Lexunit target_token_num    target_string

7           8    9                   10     11                  ...
sent_num    FE1  FE1_start:end_token FE2    FE2_start:end_token ...

multi-token targets are written 'start_end' with start = start_token_num and
end = end_token_num (hence discontinuous targets are not fully supported)
multi-token FEs are written 'start:end'
"""

import logging

import pyfn.utils.files as files_utils
import pyfn.utils.filter as filt_utils
import pyfn.utils.marshalling as marsh_utils

from pyfn.exceptions.parameter import InvalidParameterError

__all__ = ['marshall_annosets_dict']

logger = logging.getLogger(__name__)


def _get_token_num(index, text):
    token_num = -1
    prev_char_is_whitespace = True
    for char_index, char in enumerate(text):
        if char.isspace():
            prev_char_is_whitespace = True
            continue
        else:
            if prev_char_is_whitespace:
                token_num += 1
            prev_char_is_whitespace = False
        if char_index == index:
            return token_num
    raise Exception('Could not determine token number for char index '
                    '{} in sentence \'{}\''
                    .format(index, text))


def _get_fe_chunks(annoset):
    # Annosets are filtered so no start/end index points at a whitespace
    fe_chunks = ''
    if 'FE' not in annoset.labelstore.labels_by_layer_name:
        return fe_chunks
    for label in annoset.labelstore.labels_by_layer_name['FE']:
        if label.start != -1 and label.end != -1:
            start_token = _get_token_num(label.start, annoset.sentence.text)
            end_token = _get_token_num(label.end, annoset.sentence.text)
            if start_token == end_token:
                fe_chunks = '{}\t{}\t{}'.format(
                    fe_chunks, label.name, start_token).strip()
            else:
                fe_chunks = '{}\t{}\t{}:{}'.format(
                    fe_chunks, label.name, start_token, end_token).strip()
    return fe_chunks


def _get_frame_fe_num(annoset):
    frame_fe_num = 1  # at least the frame
    if 'FE' not in annoset.labelstore.labels_by_layer_name:
        return frame_fe_num
    for label in annoset.labelstore.labels_by_layer_name['FE']:
        if label.start != -1 and label.end != -1:
            frame_fe_num += 1
    return frame_fe_num


def _get_max_index(indexes):
    max_index = -1
    for (start, end) in indexes:
        if start == -1 or end == -1:
            raise InvalidParameterError('Target start/end indexes are '
                                        'undefined')
        if max_index == -1:
            max_index = max(start, end)
        else:
            max_index = max(max_index, start, end)
    return max_index


def _get_min_index(indexes):
    min_index = -1
    for (start, end) in indexes:
        if start == -1 or end == -1:
            raise InvalidParameterError('Target start/end indexes are '
                                        'undefined')
        if min_index == -1:
            min_index = min(start, end)
        else:
            min_index = min(min_index, start, end)
    return min_index


def _get_target_token_num(target, text):
    min_index = _get_min_index(target.indexes)
    max_index = _get_max_index(target.indexes)
    min_token_num = _get_token_num(min_index, text)
    max_token_num = _get_token_num(max_index, text)
    if min_token_num == max_token_num:
        return min_token_num
    return '{}_{}'.format(min_token_num, max_token_num)


def _get_semafor_line(annoset, sent_dict, train_mode):
    if not train_mode:
        return '1\t0.0\t1\t{}\t{}\t{}\t{}\t{}'.format(
            annoset.target.lexunit.frame.name,
            annoset.target.lexunit.name,
            _get_target_token_num(annoset.target, annoset.sentence.text),
            annoset.target.string,
            marsh_utils.get_sent_num(annoset.sentence.text, sent_dict))
    return '1\t0.0\t{}\t{}\t{}\t{}\t{}\t{}\t{}'.format(
        _get_frame_fe_num(annoset),
        annoset.target.lexunit.frame.name,
        annoset.target.lexunit.name,
        _get_target_token_num(annoset.target, annoset.sentence.text),
        annoset.target.string,
        marsh_utils.get_sent_num(annoset.sentence.text, sent_dict),
        _get_fe_chunks(annoset)).strip()


def _marshall_semafor(annosets, filtering_options, sent_dict,
                      semafor_filepath, excluded_frames,
                      excluded_sentences, excluded_annosets, train_mode):
    files_utils.create_parent_dir_if_not_exists(semafor_filepath)
    with open(semafor_filepath, 'w', encoding='utf-8') as semafor_stream:
        for annoset in filt_utils.filter_and_sort_annosets(annosets,
                                                           filtering_options,
                                                           excluded_frames,
                                                           excluded_sentences,
                                                           excluded_annosets):
            semafor_line = _get_semafor_line(annoset, sent_dict, train_mode)
            print(semafor_line, file=semafor_stream)


def marshall_annosets_dict(annosets_dict, target_dirpath, filtering_options,
                           output_sentences, excluded_frames,
                           excluded_sentences, excluded_annosets):
    """Convert a dict of {splits:pyfn.AnnotationSet} to SEMAFOR splits files.

    The train spits will be converted to a .frame.elements file containing
    both frame and frame element labels depending on filtering options.
    The dev/test splits will be converted to a .frames file containing
    frame labels only.
    """
    for splits_name, annosets in annosets_dict.items():
        logger.info('Marshalling {} splits to semafor format'
                    .format(splits_name))
        semafor_filepath = files_utils.get_semafor_filepath(target_dirpath,
                                                            splits_name)
        sent_filepath = files_utils.get_sent_filepath(target_dirpath,
                                                      splits_name)
        sent_dict = {}
        if splits_name not in ['train', 'dev', 'test']:
            raise InvalidParameterError('Unsupported splits_name: {}'.format(
                splits_name))
        if splits_name == 'dev' or splits_name == 'test':
            # No special filtering on dev/test
            _marshall_semafor(annosets, [], sent_dict,
                              semafor_filepath, excluded_frames,
                              excluded_sentences, excluded_annosets,
                              train_mode=False)
        elif splits_name == 'train':
            _marshall_semafor(annosets, filtering_options, sent_dict,
                              semafor_filepath, excluded_frames,
                              excluded_sentences, excluded_annosets,
                              train_mode=True)
        # print out sentences file
        if output_sentences:
            marsh_utils.marshall_sent_dict(sent_dict, sent_filepath)



"""Unmarshall FrameNet XML files."""

import os
import itertools
import logging

import xml.etree.ElementTree as etree

import pyfn.utils.constants as const
import pyfn.utils.filter as f_utils
import pyfn.utils.xml as xml_utils

from pyfn.exceptions.xml import XMLProcessingError

from pyfn.models.annotationset import AnnotationSet
from pyfn.models.corpus import Corpus
from pyfn.models.document import Document
from pyfn.models.ferelation import FERelation
from pyfn.models.frame import Frame
from pyfn.models.frameelement import FrameElement
from pyfn.models.framerelation import FrameRelation
from pyfn.models.framerelationtype import FrameRelationType
from pyfn.models.label import Label
from pyfn.models.layer import Layer
from pyfn.models.lexunit import LexUnit
from pyfn.models.sentence import Sentence

__all__ = ['extract_annosets', 'get_annosets_dict', 'extract_fn_annosets',
           'extract_relations']

logger = logging.getLogger(__name__)


def _create_label(label_tag, layer_tag):
    layer = Layer(layer_tag.get('name'))
    if layer_tag.get('rank'):
        layer.rank = int(layer_tag.get('rank'))
    label = Label(label_tag.get('name'), layer)
    if label_tag.get('start') is not None:
        label.start = int(label_tag.get('start'))
    else:
        label.start = -1
    if label_tag.get('end') is not None:
        label.end = int(label_tag.get('end'))
    else:
        label.end = -1
    if label_tag.get('feID') is not None:
        label.fe_id = int(label_tag.get('feID'))
    if label_tag.get('itype') is not None:
        label.itype = label_tag.get('itype')
    if label_tag.get('bgColor') is not None:
        label.bg_color = label_tag.get('bgColor')
    if label_tag.get('fgColor') is not None:
        label.fg_color = label_tag.get('fgColor')
    return label


def _extract_label_tags(layer_tag):
    # For FrameNet XML format
    label_tags = layer_tag.findall('fn:label', const.FN_XML_NAMESPACE)
    if label_tags:
        return label_tags
    # For SEMEVAL XML format
    labels_tags = layer_tag.findall('labels')
    if not labels_tags:
        return []
    label_tags = []
    for labels_tag in labels_tags:
        tmp_label_tags = labels_tag.findall('label')
        label_tags.extend(tmp_label_tags)
    return label_tags


def _extract_labels(layer_tags, layer_names):
    if not layer_tags:
        return []
    labels = []
    for layer_tag in layer_tags:
        if layer_tag.get('name') in layer_names:
            label_tags = _extract_label_tags(layer_tag)
            if not label_tags:
                continue
            for label_tag in label_tags:
                labels.append(_create_label(label_tag, layer_tag))
    return labels


def _extract_fn_annoset(annoset_tag, sentence, xml_schema_type,
                        annoset_layer_names, fe_dict, lexunit=None):
    _id = int(annoset_tag.get('ID'))
    logger.debug('Processing annotationSet #{}'.format(_id))
    labels = _extract_labels(_extract_layer_tags(annoset_tag),
                             annoset_layer_names)
    if lexunit is None:  # processing a fulltext file
        frame = Frame(annoset_tag.get('frameName'))
        if annoset_tag.get('frameID'):
            frame._id = int(annoset_tag.get('frameID'))
        lexunit = LexUnit(frame)
        if annoset_tag.get('luID'):
            lexunit._id = int(annoset_tag.get('luID'))
        if annoset_tag.get('luName'):
            lexunit.name = annoset_tag.get('luName')
    return AnnotationSet.from_fn_data(_id=_id, fn_labels=labels,
                                      lexunit=lexunit, sentence=sentence,
                                      fe_dict=fe_dict,
                                      xml_schema_type=xml_schema_type)


def _extract_layer_tags(annoset_tag):
    # For FrameNet XML format
    layer_tags = annoset_tag.findall('fn:layer', const.FN_XML_NAMESPACE)
    if layer_tags:
        return layer_tags
    # For SEMEVAL XML format
    layers_tags = annoset_tag.findall('layers')
    if not layers_tags:
        return []
    layer_tags = []
    for layers_tag in layers_tags:
        tmp_layer_tags = layers_tag.findall('layer')
        layer_tags.extend(tmp_layer_tags)
    return layer_tags


def _has_target_layer(annoset_tag):
    layer_tags = _extract_layer_tags(annoset_tag)
    if not layer_tags:
        return False
    for layer_tag in layer_tags:
        if layer_tag.get('name') == 'Target':
            label_tags = _extract_label_tags(layer_tag)
            if not label_tags:
                return False  # has a Target layer but no labels
            for label_tag in label_tags:
                if label_tag.get('start') is None or label_tag.get('end') is None:
                    return False
            return True
    return False


def _is_fn_annoset(annoset_tag):
    return _has_target_layer(annoset_tag)


def _extract_fn_annosets(annoset_tags, sentence, xml_schema_type,
                         annoset_layer_names, fe_dict, lexunit=None):
    return [_extract_fn_annoset(annoset_tag, sentence, xml_schema_type,
                                annoset_layer_names,
                                fe_dict=fe_dict, lexunit=lexunit)
            for annoset_tag in annoset_tags
            if _is_fn_annoset(annoset_tag)]


def _extract_sentence_text(sentence_tag):
    text_tag = sentence_tag.find('fn:text', const.FN_XML_NAMESPACE)
    # Careful: in gold FN data, some lines are
    # not trimmed. Ex: '_whitespace_ Simply put , Stephanopoulos did
    # as much...' hence the rstrip and not a simple strip
    if text_tag is not None:
        return text_tag.text.rstrip()
    return sentence_tag.find('text').text.rstrip()


def _extract_sentence(sentence_tag, pnwb_labels, document=None):
    sentence_text = _extract_sentence_text(sentence_tag)
    sentence = Sentence(text=sentence_text, _id=int(sentence_tag.get('ID')),
                        pnwb_labels=pnwb_labels)
    logger.debug('Processing sentence #{}: {}'.format(sentence._id,
                                                      sentence.text))
    if document:
        sentence.document = document
    return sentence


def _extract_sent_labels(annoset_tags, layer_names):
    all_labels = []
    for annoset_tag in annoset_tags:
        layer_tags = _extract_layer_tags(annoset_tag)
        all_labels.extend(_extract_labels(layer_tags, layer_names))
    return all_labels


def _extract_annoset_tags(sentence_tag):
    # For FrameNet XML format
    annoset_tags = sentence_tag.findall('fn:annotationSet',
                                        const.FN_XML_NAMESPACE)
    if annoset_tags:
        return annoset_tags
    # For SEMEVAL XML format
    annosets_tags = sentence_tag.findall('annotationSets')
    if not annosets_tags:
        return []
    annoset_tags = []
    for annosets_tag in annosets_tags:
        tmp_annoset_tags = annosets_tag.findall('annotationSet')
        annoset_tags.extend(tmp_annoset_tags)
    return annoset_tags


def extract_fn_annosets(sentence_tag, xml_schema_type, fe_dict,
                        document=None, lexunit=None):
    """Return a [AnnotationSet,...,] extracted from a single <sentence> tag."""
    annoset_tags = _extract_annoset_tags(sentence_tag)
    if not annoset_tags:
        return []
    logger.debug('Processing {} annotationSet tags'.format(len(annoset_tags)))
    pnwb_labels = _extract_sent_labels(annoset_tags, const.SENT_LAYERS)
    sentence = _extract_sentence(sentence_tag, pnwb_labels, document=document)
    return _extract_fn_annosets(annoset_tags, sentence, xml_schema_type,
                                const.ANNO_LAYERS,
                                fe_dict=fe_dict, lexunit=lexunit)


def _unmarshall_exemplar_xml(xml_file_path, fe_dict, flatten=False):
    """Unmarshall a FrameNet lu XML file from file path.

    Return a generator over a list of AnnotationSet instances extracted
    from the lu XML file. A single list of AnnotationSet instances corresponds
    to all the AnnotationSets of a single Sentence.

    Args
    ----
        param1 xml_files_path: full path to a FrameNet lu XML file.

    """
    logger.info('Unmarshalling FrameNet lu XML file: {}'.format(xml_file_path))
    tree = etree.parse(xml_file_path)
    root = tree.getroot()
    frame = Frame(root.get('frame'), _id=int(root.get('frameID')))
    lexunit = LexUnit(frame, _id=int(root.get('ID')), name=root.get('name'))
    subcorpus_tags = root.findall('fn:subCorpus', const.FN_XML_NAMESPACE)
    for subcorpus_tag in subcorpus_tags:
        sentence_tags = subcorpus_tag.findall('fn:sentence',
                                              const.FN_XML_NAMESPACE)
        for sentence_tag in sentence_tags:
            annosets = extract_fn_annosets(
                sentence_tag, xml_schema_type='exemplar', fe_dict=fe_dict,
                lexunit=lexunit)
            if annosets:
                if not flatten:
                    yield annosets
                if flatten:
                    for annoset in annosets:
                        yield annoset


def _extract_document(header_tag):
    """Return a Document instance from a fulltext <header> tag."""
    corpus_tag = header_tag.find('fn:corpus', const.FN_XML_NAMESPACE)
    corpus = Corpus(int(corpus_tag.get('ID')), corpus_tag.get('name'),
                    corpus_tag.get('description'))
    document_tag = corpus_tag.find('fn:document', const.FN_XML_NAMESPACE)
    if document_tag is None:
        raise XMLProcessingError('Could not extract document from tag')
    return Document(int(document_tag.get('ID')),
                    document_tag.get('name'),
                    document_tag.get('description'),
                    corpus)


def _unmarshall_fulltext_xml(xml_file_path, fe_dict, flatten=False):
    """Unmarshall a FrameNet fulltext XML file from file path.

    Return a generator on AnnotationSet instances extracted from the
    fulltext XML file. A single list of AnnotationSet instances corresponds
    to all the AnnotationSets of a single Sentence.

    Args
    ----
        param1 xml_files_path: full path to a FrameNet fulltext XML file.

    """
    logger.info('Unmarshalling FrameNet fulltext XML file: {}'.format(
        xml_file_path))
    tree = etree.parse(xml_file_path)
    root = tree.getroot()
    try:
        document = _extract_document(root.find('fn:header',
                                               const.FN_XML_NAMESPACE))
    except XMLProcessingError as err:
        raise XMLProcessingError('Could not process XML file: {}. Cause: {}'
                                 .format(xml_file_path, str(err)))
    sentence_tags = root.findall('fn:sentence', const.FN_XML_NAMESPACE)
    for sentence_tag in sentence_tags:
        annosets = extract_fn_annosets(
            sentence_tag, xml_schema_type='fulltext', fe_dict=fe_dict,
            document=document)
        if annosets:
            if not flatten:
                yield annosets
            if flatten:
                for annoset in annosets:
                    yield annoset


def _extract_ex_annosets(ex_filepaths, fe_dict, flatten=False):
    generators = []
    for exemplar_filepath in ex_filepaths:
        generators.append(_unmarshall_exemplar_xml(
            exemplar_filepath, fe_dict, flatten))
    return generators


def _extract_ft_annosets(ft_filepaths, fe_dict, flatten=False):
    generators = []
    for fulltext_filepath in ft_filepaths:
        generators.append(_unmarshall_fulltext_xml(
            fulltext_filepath, fe_dict, flatten))
    return generators


def extract_annosets(splits_dirpath, with_fulltexts, with_exemplars,
                     fe_dict, flatten=False):
    """Return a list of pyfn.AnnotationSet extracted from splits paths.

    The splits directory should contain two subdirectories name 'fulltext'
    and 'lu'.
    Returns a list of generators over AnnotationSet with [AnnotationSet] <->
    1 fulltext or exemplar XML file if flatten==False
    Returns a generator over AnnotationSet (a single list) if flatten==True
    """
    logger.info('Extracting pyfn.AnnotationSet items from {}'
                .format(splits_dirpath))
    ft_annosets = []
    ex_annosets = []
    if with_fulltexts:
        ft_filepaths = xml_utils.get_xml_filepaths(splits_dirpath,
                                                   'fulltext')
        ft_annosets = _extract_ft_annosets(ft_filepaths, fe_dict, flatten)
    if with_exemplars:
        ex_filepaths = xml_utils.get_xml_filepaths(splits_dirpath,
                                                   'lu')
        ex_annosets = _extract_ex_annosets(ex_filepaths, fe_dict, flatten)
    return itertools.chain(*ft_annosets, *ex_annosets)


def _filter_annosets_dict(annosets_dict):
    # At least test and train, dev is optional
    filtered_annosets_dict = {}
    test_annosets, _test_annosets, __test_annosets = itertools.tee(
        annosets_dict['test'], 3)
    filtered_annosets_dict['test'] = test_annosets
    if annosets_dict['dev']:
        filtered_dev_annosets = f_utils.left_difference(
            annosets_dict['dev'], _test_annosets)
        dev_annosets, _dev_annosets = itertools.tee(filtered_dev_annosets)
        filtered_annosets_dict['dev'] = dev_annosets
        filtered_annosets_dict['train'] = f_utils.left_difference(
            annosets_dict['train'], itertools.chain(_dev_annosets,
                                                    __test_annosets))
    else:
        filtered_annosets_dict['train'] = f_utils.left_difference(
            annosets_dict['train'], _test_annosets)
    return filtered_annosets_dict


def _extract_frame_element(fe_tag):
    return FrameElement(_id=int(fe_tag.get('ID')), name=fe_tag.get('name'),
                        coretype=fe_tag.get('coreType'))


def _get_fe_dict(frame_xml_filepaths):
    fe_dict = {}
    if not frame_xml_filepaths:
        logger.warning('Could not find \'frame\' directory under {}. '
                       'Unmarshalling FrameNet XML files without '
                       'FrameElement dictionary. '
                       'If you wish to access FE coreType information, '
                       'you may want to start over after adding the FrameNet '
                       'frame/ subdirectory containing frame XML files '
                       'under your FN splits directory'.format(
                           frame_xml_filepaths))
        return fe_dict
    for frame_xml_filepath in frame_xml_filepaths:
        logger.info('Unmarshalling frame XML file {}'.format(frame_xml_filepath))
        tree = etree.parse(frame_xml_filepath)
        root = tree.getroot()
        try:
            fe_tags = root.findall('fn:FE', const.FN_XML_NAMESPACE)
            for fe_tag in fe_tags:
                frame_element = _extract_frame_element(fe_tag)
                fe_dict[frame_element._id] = frame_element
        except XMLProcessingError as err:
            raise XMLProcessingError('Could not process XML file: {}. Cause: {}'
                                     .format(frame_xml_filepath, str(err)))
    return fe_dict


def _get_annosets_dict_from_fn_xml(fn_splits_dirpath, splits, with_exemplars):
    fe_dict = _get_fe_dict(xml_utils.get_xml_filepaths(fn_splits_dirpath,
                                                       'frame'))
    if splits == 'test':
        return {'test': extract_annosets(os.path.join(fn_splits_dirpath,
                                                      'test'),
                                         with_fulltexts=True,
                                         with_exemplars=with_exemplars,
                                         fe_dict=fe_dict,
                                         flatten=True),
                'dev': [], 'train': []}
    if splits == 'dev':
        return {'test': extract_annosets(os.path.join(fn_splits_dirpath,
                                                      'test'),
                                         with_fulltexts=True,
                                         with_exemplars=with_exemplars,
                                         fe_dict=fe_dict,
                                         flatten=True),
                'dev': extract_annosets(os.path.join(fn_splits_dirpath,
                                                     'dev'),
                                        with_fulltexts=True,
                                        with_exemplars=with_exemplars,
                                        fe_dict=fe_dict,
                                        flatten=True),
                'train': []}
    if splits == 'train':
        return {'test': extract_annosets(os.path.join(fn_splits_dirpath,
                                                      'test'),
                                         with_fulltexts=True,
                                         with_exemplars=with_exemplars,
                                         fe_dict=fe_dict,
                                         flatten=True),
                'dev': extract_annosets(os.path.join(fn_splits_dirpath,
                                                     'dev'),
                                        with_fulltexts=True,
                                        with_exemplars=with_exemplars,
                                        fe_dict=fe_dict,
                                        flatten=True),
                'train': extract_annosets(os.path.join(fn_splits_dirpath,
                                                       'train'),
                                          with_fulltexts=True,
                                          with_exemplars=with_exemplars,
                                          fe_dict=fe_dict,
                                          flatten=True)}
    return {}


def get_annosets_dict(source_path, splits, with_exemplars):
    """Return a string to AnnotationSet generator dict.

    Keys are splits name (train, dev, test) and values are generators
    on AnnotationSet objects.
    """
    logger.info('Creating pyfn.AnnotationSet dict from {}'.format(source_path))
    return _filter_annosets_dict(
        _get_annosets_dict_from_fn_xml(source_path, splits, with_exemplars))


def _extract_fe_relation(fe_relation_tag, frame_relation):
    return FERelation(
        _id=int(),
        sub_fe=FrameElement(_id=int(fe_relation_tag.get('subID')),
                            name=fe_relation_tag.get('subFEName')),
        sup_fe=FrameElement(_id=int(fe_relation_tag.get('supID')),
                            name=fe_relation_tag.get('superFEName')),
        frame_relation=frame_relation)


def _extract_frame_relation(frame_relation_tag, frame_relation_type):
    return FrameRelation(
        _id=int(frame_relation_tag.get('ID')),
        sub_frame=Frame(_id=int(frame_relation_tag.get('subID')),
                        name=frame_relation_tag.get('subFrameName')),
        sup_frame=Frame(_id=int(frame_relation_tag.get('supID')),
                        name=frame_relation_tag.get('superFrameName')),
        frtype=frame_relation_type)


def _extract_frame_relation_type(fr_type_tag):
    return FrameRelationType(
        _id=int(fr_type_tag.get('ID')), name=fr_type_tag.get('name'),
        sub_frame_name=fr_type_tag.get('subFrameName'),
        sup_frame_name=fr_type_tag.get('superFrameName'))


def extract_relations(fr_relation_xml_filepath):
    """Extract a (List<FrameRelation>, List<FERelation>) tuple.

    Tuple is extracted from a FrameNet frRelation.xml file.
    """
    frame_relations = []
    fe_relations = []
    tree = etree.parse(fr_relation_xml_filepath)
    root = tree.getroot()
    try:
        fr_types_tags = root.findall('fn:frameRelationType',
                                     const.FN_XML_NAMESPACE)
        for fr_type_tag in fr_types_tags:
            frame_relation_type = _extract_frame_relation_type(fr_type_tag)
            frame_relation_tags = fr_type_tag.findall('fn:frameRelation',
                                                      const.FN_XML_NAMESPACE)
            for frame_relation_tag in frame_relation_tags:
                frame_relation = _extract_frame_relation(frame_relation_tag,
                                                         frame_relation_type)
                frame_relations.append(frame_relation)
                fe_relation_tags = frame_relation_tag.findall(
                    'fn:FERelation', const.FN_XML_NAMESPACE)
                for fe_relation_tag in fe_relation_tags:
                    fe_relation = _extract_fe_relation(fe_relation_tag,
                                                       frame_relation)
                    fe_relations.append(fe_relation)
    except XMLProcessingError as err:
        raise XMLProcessingError('Could not process XML file: {}. Cause: {}'
                                 .format(fr_relation_xml_filepath, str(err)))
    return frame_relations, fe_relations
